(* Authors:  Christian Sternagel and Bertram Felgenhauer *)

theory Insertion_Sort
imports Main
begin

section \<open>Insertion sort\<close>

subsection \<open>Implementation\<close>

fun insert :: "'a::linorder \<Rightarrow> 'a list \<Rightarrow> 'a list"
where
  "insert x [] = [x]"
| "insert x (y # ys) = (if x \<le> y then x # y # ys else y # insert x ys)"

abbreviation "isort xs \<equiv> fold insert xs []"

subsection \<open>Correctness\<close>

text \<open>@{term "isort"} produces sorted lists\<close>

lemma set_insert [simp]:
  "set (insert x xs) = {x} \<union> set xs"
by (induct xs) auto

lemma sorted_insert:
  "sorted ys \<Longrightarrow> sorted (insert x ys)"
by (induct ys) (auto simp: sorted_Cons)

lemma sorted_isort_append:
  "\<forall>ys. sorted ys \<longrightarrow> sorted (fold insert xs ys)"
proof (induct xs)
  case Nil then show ?case by simp
next
  case (Cons x xs) then show ?case
  by (auto intro: sorted_insert)
qed

lemma sorted_isort: "sorted (fold insert xs [])"
using sorted_isort_append[of xs] by simp

text \<open>@{term "isort"} preserves the elements of the list\<close>

fun count :: "'a \<Rightarrow> 'a list \<Rightarrow> nat"
where
  "count x [] = 0"
| "count x (y # ys) = (if x = y then count x ys + 1 else count x ys)"

lemma count_insert [simp]:
  "count x (insert y ys) = (if x = y then count x ys + 1 else count x ys)"
by (induct ys) auto

lemma count_fold_insert [simp]:
  "\<forall>ys. count x (fold insert xs ys) = count x xs + count x ys"
by (induct xs) auto

text \<open>@{term "isort"} preserves the elements of the list\<close>

lemma count_isort: "count x (isort xs) = count x xs" by simp

end
